{$B-,H+,J+,Q-,T-,X+}

{$UNDEF D3PLUS}
{$UNDEF D4PLUS}
{$IFDEF VER100}{$DEFINE D3PLUS}{$ENDIF}
{$IFDEF VER120}{$DEFINE D3PLUS}{$DEFINE D4PLUS}{$ENDIF}
{$IFDEF VER130}{$DEFINE D3PLUS}{$DEFINE D4PLUS}{$ENDIF}

unit GpTextFile;

(*:Interface to 8/16-bit text files. Uses GpHugeF unit for file access.
   @author Primoz Gabrijelcic
   @desc <pre>
   (c) 2000 Primoz Gabrijelcic
   Free for personal and commercial use. No rights reserved.

   Author           : Primoz Gabrijelcic
   Creation date    : 1999-11-01
   Last modification: 2001-02-27
   Version          : 2.05
   Requires         : GpHugeF 3.01
   </pre>
*)(*
   History:
     2.05: 2001-02-27
       - Added property AcceptedDelimiters to TGpTextFile and TGpTextFileStream.
         You can use it do specify what CR-LF combinations should be treated as
         a line delimiter. 
         If not set (or set to []), classes will behave as before - TGpTextFile
         will use CR, LF, CRLF, /2028/, and /000D/000A/ for line delimiters and
         TGpTextFileStream will use LF, CRLF, /2028/, and /000D/000A/ for line
         delimiters.

     2.04: 2001-01-31
       - All raised exceptions now have HelpContext set. All possible
         HelpContext values are enumerated in 'const' section at the very
         beginning of the unit. Thanks to Peter Evans for the suggestion.
       - TGpTextFileStream.Win32Check made protected (was public by mistake).

     2.03a: 2000-12-11
       - Fixed Append and AppendSafe to create a file if it does not exist.

     2.03: 2000-10-25
       - Removed one FreeAndNil call preventing unit to compile under Delphi 2,
         3, and 4.

     2.02: 2000-10-21
       - Added support for Unicode pseudo-codepage 1200 (CP_UNICODE).

     2.01: 2000-10-12
       - Added UTF-8 support (CP_UTF8 codepage) to TGpTextFile and
         TGpTextFileStream. To work with Unicode file in UTF-8 encoding, set
         cfUnicode flag and CP_UTF8 code page at the same time.
         When working in UTF-8 mode, TGpTextFile doesn't support lines longer
         than 2.147.483.647 bytes.
         - When opening files in unknown format, it is probably best to first
           Reset it with flags=[cfUnicode] and codepage=CP_UTF8. Then you can
           check if file is 16-bit Unicode (Is16bit). If it is not, you can
           change the Codepage to some other value (1252, for example) or just
           leave it at CP_UTF8 (reasonable default for processing XML files).
       - UTF-8 support will read and skip Byte Order Mark (EF BB BF) but will
         only write BOM if cfWriteUTF8BOM is specified in flags.
       - New functions TGpTextFile.Is16bit and TGpTextFileStream.Is16bit return
         true when file is Unicode with 16-bit (UCS-2) encoding. IsUnicode
         returns true when file is Unicode with 16-bit (UCS-2) or 8-bit (UTF-8)
         encoding.
       - Added symbolic constants for ISO code pages not defined in Windows.pas:
         ISO-8859-1, ISO-8859-2, ISO-8859-3, ISO-8859-4, ISO-8859-5, ISO-8859-6,
         ISO-8859-7, ISO-8859-8 (symbols are named ISO_8859_n).

     2.0: 2000-10-06
       - Created TGpTextFileStream class, descendant of TStream, which offers
         similar functionality as TGpTextFile - automatic detection of Unicode
         files, automatic code page remapping etc.
       - TGpTextFile.SetCodepage made public so it is possible to change code
         page on the fly.
       - Added parameters 'flags', 'waitObject', and 'codePage' to
         TGpTextFile.AppendSafe.
       - Order of 'flags' and 'bufferSize' parameters reversed in most methods.
       - Position of 'diskLockTimeout' and 'diskRetryDelay' parameters changed
         in all methods. 
       - Fully documented.
       - All language-dependant string constants moved to resourcestring section.

     1.08: 2000-10-05
       - Added codePage parameter to all Reset and Rewrite functions. Default
         value of 0 specifies conversion according to current codepage and any
         other number specifies codepage that should be used for conversion.

     1.07: 2000-08-01
       - All exceptions generated in this unit were converted to EGpTextFile
         exceptions (descendant of EGpHugeFile).
       - All Windows-generated exceptions not caught in TGpHugeFile are now
         converted to EGpTextFile exception.
       - Added parameter bufferSize to Append and AppendSafe.
       - Append* opened file in non-buffered mode. Fixed.

     1.06: 2000-06-22
       - Added overloaded version of Writeln.

     1.05: 2000-05-15
       - Rewrite now opens file in buffered mode.

     1.04: 2000-04-19
       - Added new ResetSafe/RewriteSafe parameter - waitObject. It is forwarded to
         TGpHugeFile.ResetEx/RewriteEx.

     1.03: 2000-03-03
       - Added OpenFlags parameter to Reset and ResetSafe. In D4 or higher its
         value defaults to [] so no source code changes to old applications will
         be required. Currently only supported OpenFlag is opCloseOnEOF, which
         enables hfoCloseOnEOF flag in TGpHugeFile (see GpHugeF.pas for more
         details).

     1.02: 1999-11-24
       - Added bufferSize parameter to Reset* and Rewrite*. By default
         (bufferSize = 0), 64 KB buffer is allocated.

     1.01a: 1999-11-03
       - Append fixed.

     1.01: 1999-11-02
       - Added ResetSafe and RewriteSafe methods.

     1.0: 1999-11-01
       - First published version.
*)

interface

uses
  Windows,
  Classes,
  GpHugeF;

// HelpContext values for all raised exceptions. All EGpHugeFile exception are
// re-raised without modifying HelpContext (which was already assigned in
// GpHugeF unit).
const
  //:Exception was handled and converted to EGpTextFile but was not expected and is not categorised.
  hcTFUnexpected              = 2000;
  //:Windows error.
  hcTFWindowsError            = 2001;
  //:Unknown Windows error.
  hcTFUnknownWindowsError     = 2002;
  //:Failed to append file.
  hcTFFailedToAppend          = 2003;
  //:Failed to reset file.
  hcTFFailedToReset           = 2004;
  //:Failed to rewrite file.
  hcTFFailedToRewrite         = 2005;
  //:Cannot append reversed Unicode file - not supported.
  hcTFCannotAppendReversed    = 2006;
  //:Cannot write to reversed Unicode file - not supported.
  hcTFCannotWriteReversed     = 2007;
  //:Parameter to Write method is invalid.
  hcTFInvalidParameter        = 2008;
  //:Cannot convert odd number of bytes.
  hcTFCannotConvertOdd        = 2009;

const
{$IFNDEF D3plus}
  CP_UTF8 = 65001;    // UTF-8 pseudo-codepage, defined in Windows.pas in Delphi 3 and newer.
{$ENDIF}

  CP_UNICODE =  1200; // Unicode pseudo-codepage,
  ISO_8859_1 = 28591; // Western Alphabet (ISO)
  ISO_8859_2 = 28592; // Central European Alphabet (ISO)
  ISO_8859_3 = 28593; // Latin 3 Alphabet (ISO)
  ISO_8859_4 = 28594; // Baltic Alphabet (ISO)
  ISO_8859_5 = 28595; // Cyrillic Alphabet (ISO)
  ISO_8859_6 = 28596; // Arabic Alphabet (ISO)
  ISO_8859_7 = 28597; // Greek Alphabet (ISO)
  ISO_8859_8 = 28598; // Hebrew Alphabet (ISO)

type
  {:Base exception class for exceptions created in TGpTextFile and descendants.
  }
  EGpTextFile       = class(EGpHugeFile);

  {:Base exception class for exceptions created in TGpTextFileStream.
  }
  EGpTextFileStream = class(EGpHugeFileStream);

  {:Text file creation flags.
    @enum cfUnicode          Create Unicode file.
    @enum cfReverseByteOrder Create unicode file with reversed byte order
                             (Motorola format). Set only on Reset, readonly.
                             Currently ignored in Rewrite.
    @enum cfUse2028          Use standard /2028/ instead of /000D/000A/ for line
                             delimiter (MS Notepad and MS Word do not understand
                             $2028 delimiter). Applies to Unicode files only.
    @enum cfUseLF            Use /LF/ instead of /CR/LF/ for line delimiter.
                             Applies to 8-bit files only.
    @enum cfWriteUTF8BOM     Write UTF-8 Byte Order Mark to the beginning of
                             file.
  }
  TCreateFlag = (cfUnicode, cfReverseByteOrder, cfUse2028, cfUseLF,
    cfWriteUTF8BOM);

  {:Set of all creation flags.
  }
  TCreateFlags = set of TCreateFlag;

  {:Text file open (reset) flags.
    @enum ofCloseOnEOF Remaps to TGpHugeFile hfCloseOnEOF.
  }
  TOpenFlag = (ofCloseOnEOF);

  {:Set of all open flags.
  }
  TOpenFlags = set of TOpenFlag;

  {:Line delimiters.
    @enum ldCR       Carriage return (Mac style).
    @enum ldLF       Line feed (Unix style).
    @enum ldCRLF     Carriage return + Line feed (DOS style).
    @enum ldLFCR     Line feed + Carriage return (very unusual combination).
    @enum ld2028     /2028/ Unicode delimiter.
    @enum ld000D000A /000D/000A/ Windows-style Unicode delimiter.
  }
  TLineDelimiter = (ldCR, ldLF, ldCRLF, ldLFCR, ld2028, ld000D000A);

  {:Set of all line delimiters.
  }
  TLineDelimiters = set of TLineDelimiter;

  {:Unified 8/16-bit text file access. All strings passed as Unicode, conversion
    to/from 8-bit is done automatically according to specified code page.
    Access is buffered but direct-access functions (FilePos, Seek) are supported
    nevertheless.
  }
  TGpTextFile = class(TGpHugeFile)
  private
    tfCFlags    : TCreateFlags;
    tfCodePage  : word;
    tfLineDelims: TLineDelimiters;
    tfSmallBuf  : pointer;
  protected
    function  AllocTmpBuffer(size: integer): pointer; virtual;
    procedure FreeTmpBuffer(var buffer: pointer); virtual;
    procedure PrepareBuffer; virtual;
    procedure SetCodepage(cp: word); virtual;
    procedure WriteString(ws: WideString); virtual;
  public
    destructor Destroy; override;
    procedure Append(
      flags: TCreateFlags {$IFDEF D4plus}= []{$ENDIF};
      bufferSize: integer {$IFDEF D4plus}= 0{$ENDIF};
      codePage: word      {$IFDEF D4plus}= 0{$ENDIF});
    function  AppendSafe(
      flags: TCreateFlags      {$IFDEF D4plus}= []{$ENDIF};
      bufferSize: integer      {$IFDEF D4plus}= 0{$ENDIF};
      diskLockTimeout: integer {$IFDEF D4plus}= 0{$ENDIF};
      diskRetryDelay: integer  {$IFDEF D4plus}= 0{$ENDIF};
      waitObject: THandle      {$IFDEF D4plus}= 0{$ENDIF};
      codePage: word           {$IFDEF D4plus}= 0{$ENDIF}): THFError;
    function  EOF: boolean;
    function  Is16bit: boolean;
    function  IsUnicode: boolean;
    function  Readln: WideString;
    procedure Reset(
      flags: TOpenFlags   {$IFDEF D4plus}= []{$ENDIF};
      bufferSize: integer {$IFDEF D4plus}= 0{$ENDIF};
      codePage: word      {$IFDEF D4plus}= 0{$ENDIF});
    function  ResetSafe(
      flags: TOpenFlags        {$IFDEF D4plus}= []{$ENDIF};
      bufferSize: integer      {$IFDEF D4plus}= 0{$ENDIF};
      diskLockTimeout: integer {$IFDEF D4plus}= 0{$ENDIF};
      diskRetryDelay: integer  {$IFDEF D4plus}= 0{$ENDIF};
      waitObject: THandle      {$IFDEF D4plus}= 0{$ENDIF};
      codePage: word           {$IFDEF D4plus}= 0{$ENDIF}): THFError;
    procedure Rewrite(
      flags: TCreateFlags {$IFDEF D4plus}= []{$ENDIF};
      bufferSize: integer {$IFDEF D4plus}= 0{$ENDIF};
      codePage: word      {$IFDEF D4plus}= 0{$ENDIF});
    function  RewriteSafe(
      flags: TCreateFlags      {$IFDEF D4plus}= []{$ENDIF};
      bufferSize: integer      {$IFDEF D4plus}= 0{$ENDIF};
      diskLockTimeout: integer {$IFDEF D4plus}= 0{$ENDIF};
      diskRetryDelay: integer  {$IFDEF D4plus}= 0{$ENDIF};
      waitObject: THandle      {$IFDEF D4plus}= 0{$ENDIF};
      codePage: word           {$IFDEF D4plus}= 0{$ENDIF}): THFError;
    procedure Write(params: array of const);
    procedure Writeln(ln: WideString {$IFDEF D4plus}= ''{$ENDIF}); {$IFDEF D4plus} overload;
    procedure Writeln(params: array of const); overload; {$ENDIF D4plus}
    {:Accepted line delimiters (CR, LF or any combination).
    }
    property AcceptedDelimiters: TLineDelimiters read tfLineDelims
      write tfLineDelims;
    {:Code page used to convert 8-bit files to Unicode and back. May be changed
      while file is open (and even partially read). If set to 0, current default
      code page will be used.
    }
    property Codepage: word read tfCodepage write SetCodepage;
  end; { TGpTextFile }

  {:Unified 8/16-bit text stream access. All strings passed as Unicode,
    conversion to/from 8-bit is done automatically according to specified code
    page. Access is buffered.
  }
  TGpTextFileStream = class(TStream)
  private
    tfsAccess      : TGpHugeFileStreamAccess;
    tfsCodePage    : word;
    tfsCreateFlags : TCreateFlags;
    tfsExternalHFS : boolean;
    tfsLineDelims  : TLineDelimiters;
    tfsOpenFlags   : TOpenFlags;
    tfsReadlnBuf   : TMemoryStream;
    tfsSmallBuf    : pointer;
    tfsStream      : TGpHugeFileStream;
    tfsWindowsError: DWORD;
  protected
    function  AllocBuffer(size: integer): pointer; virtual;
    procedure FreeBuffer(var buffer: pointer); virtual;
    function  GetFileName: string; virtual;
    function  GetSize: longint; virtual;
    function  GetWindowsError: DWORD; virtual;
    procedure PrepareStream; virtual;
    procedure SetCodepage(cp: word); virtual;
    procedure SetSize(newSize: longint); override;
    procedure Win32Check(condition: boolean; method: string);
  public
    constructor Create(
      const fileName: string; access: TGpHugeFileStreamAccess;
      openFlags: TOpenFlags     {$IFDEF D4plus}= []{$ENDIF};
      createFlags: TCreateFlags {$IFDEF D4plus}= []{$ENDIF};
      codePage: word            {$IFDEF D4plus}= 0{$ENDIF}
      );
    constructor CreateFromHandle(
      hfs: TGpHugeFileStream; access: TGpHugeFileStreamAccess;
      openFlags: TOpenFlags     {$IFDEF D4plus}= []{$ENDIF};
      createFlags: TCreateFlags {$IFDEF D4plus}= []{$ENDIF};
      codePage: word            {$IFDEF D4plus}= 0{$ENDIF}
      ); {$IFDEF D4plus}overload;{$ENDIF}
    destructor  Destroy; override;
    function  Is16bit: boolean;
    function  IsUnicode: boolean;
    function  Read(var buffer; count: longint): longint; override;
    function  Readln: WideString;
    function  Seek(offset: longint; mode: word): longint; override;
    function  Write(const buffer; count: longint): longint; override;
    function  Writeln(const ln: WideString{$IFDEF D4plus}= ''{$ENDIF}): boolean;
    function  WriteString(const ws: WideString): boolean;
    {:Accepted line delimiters (CR, LF or any combination).
    }
    property AcceptedDelimiters: TLineDelimiters read tfsLineDelims
      write tfsLineDelims;
    {:Code page used to convert 8-bit files to Unicode and back. May be changed
      while file is open (and even partially read). If set to 0, current default
      code page will be used.
    }
    property  Codepage: word read tfsCodePage write SetCodepage;
    {:Name of underlying file.
    }
    property  FileName: string read GetFileName;
    {:Stream size. Reintroduced to override GetSize (static in TStream) with
      faster version.
    }
    property  Size: longint read GetSize write SetSize;
    {:Last Windows error code.
    }
    property  WindowsError: DWORD read GetWindowsError;
  end; { TGpTextFileStream }

implementation

uses
  SysUtils,
  SysConst;

const
  {:Header for 'normal' Unicode file (Intel format).
  }
  CUnicodeNormal  : WideChar = WideChar($FEFF);

  {:Header for 'reversed' Unicode file (Motorola format).
  }
  CUnicodeReversed: WideChar = WideChar($FFFE);

  {:First two bytes of UTF-8 BOM.
  }
  CUTF8BOM12: WideChar = WideChar($BBEF);

  {:Third byte of UTF-8 BOM.
  }
  CUTF8BOM3: Char = Char($BF);

  {:Size of preallocated buffer used for 8 to 16 to 8 bit conversions in
    TGpTextFileStream.
  }
  CTFSSmallBufSize = 2048; // 1024 WideChars

{$IFDEF D3plus}
resourcestring
{$ELSE}
const
{$ENDIF}
  sCannotAppendReversedUnicodeFile   = 'TGpTextFile(%s):Cannot append reversed Unicode file.';
  sCannotAppendReversedUnicodeStream = 'TGpTextFileStream(%s):Cannot append reversed Unicode file.';
  sCannotConvertOddNumberOfBytes     = 'TGpTextFileStream(%s):Cannot convert odd number of bytes: %d';
  sCannotWriteReversedUnicodeFile    = 'TGpTextFile(%s):Cannot write to reversed Unicode file.';
  sCannotWriteReversedUnicodeStream  = 'TGpTextFileStream(%s):Cannot write to reversed Unicode file.';
  sFailedToAppendFile                = 'TGpTextFile(%s):Failed to append.';
  sFailedToResetFile                 = 'TGpTextFile(%s):Failed to reset file.';
  sFailedToRewriteFile               = 'TGpTextFile(%s):Failed to rewrite file.';
  sInvalidParameter                  = 'TGpTextFile(%s):Invalid parameter!';
  sStreamFailed                      = 'TGpTextFileStream.%s(%s) failed. ';

{:Converts Ansi string to Unicode string using specified code page.
  @param   s        Ansi string.
  @param   codePage Code page to be used in conversion.
  @returns Converted wide string.
}
function StringToWideString(const s: AnsiString; codePage: word): WideString;
var
  l: integer;
begin
  if s = '' then
    Result := ''
  else begin
    l := MultiByteToWideChar(codePage, MB_PRECOMPOSED, PChar(@s[1]), -1, nil, 0);
    SetLength(Result, l-1);
    if l > 1 then
      MultiByteToWideChar(CodePage, MB_PRECOMPOSED, PChar(@s[1]), -1, PWideChar(@Result[1]), l-1);
  end;
end; { StringToWideString }

{:Converts Unicode string to Ansi string using specified code page.
  @param   ws       Unicode string.
  @param   codePage Code page to be used in conversion.
  @returns Converted ansi string.
}
function WideStringToString (const ws: WideString; codePage: Word): AnsiString;
var
  l: integer;
begin
  if ws = '' then
    Result := ''
  else begin
    l := WideCharToMultiByte(codePage,
           WC_COMPOSITECHECK or WC_DISCARDNS or WC_SEPCHARS or WC_DEFAULTCHAR,
           @ws[1], -1, nil, 0, nil, nil);
    SetLength(Result, l-1);
    if l > 1 then
      WideCharToMultiByte(codePage,
        WC_COMPOSITECHECK or WC_DISCARDNS or WC_SEPCHARS or WC_DEFAULTCHAR,
        @ws[1], -1, @Result[1], l-1, nil, nil);
  end;
end; { WideStringToString }

{:Convers buffer of WideChars into UTF-8 encoded form. Target buffer must be
  pre-allocated and large enough (each WideChar will use at most three bytes
  in UTF-8 encoding).                                                            <br>
  RFC 2279 (http://www.ietf.org/rfc/rfc2279.txt) describes the conversion:       <br>
  $0000..$007F => $00..$7F                                                       <br>
  $0080..$07FF => 110[bit10..bit6] 10[bit5..bit0]                                <br>
  $0800..$FFFF => 1110[bit15..bit12] 10[bit11..bit6] 10[bit5..bit0]
  @param   unicodeBuf   Buffer of WideChars.
  @param   uniByteCount Size of unicodeBuf, in bytes.
  @param   utf8Buf      Pre-allocated buffer for UTF-8 encoded result.
  @returns Number of bytes used in utf8Buf buffer.
  @since   2.01
}
function WideCharBufToUTF8Buf(const unicodeBuf; uniByteCount: integer;
  var utf8Buf): integer;
var
  iwc: integer;
  pch: PChar;
  pwc: PWideChar;
  wc : word;

  procedure AddByte(b: byte);
  begin
    pch^ := char(b);
    Inc(pch);
  end; { AddByte }

begin { WideCharBufToUTF8Buf }
  pwc := @unicodeBuf;
  pch := @utf8Buf;
  for iwc := 1 to uniByteCount div SizeOf(WideChar) do begin
    wc := Ord(pwc^);
    Inc(pwc);
    if (wc >= $0001) and (wc <= $007F) then begin
      AddByte(wc AND $7F);
    end
    else if (wc >= $0080) and (wc <= $07FF) then begin
      AddByte($C0 OR ((wc SHR 6) AND $1F));
      AddByte($80 OR (wc AND $3F));
    end
    else begin // (wc >= $0800) and (wc <= $FFFF)
      AddByte($E0 OR ((wc SHR 12) AND $0F));
      AddByte($80 OR ((wc SHR 6) AND $3F));
      AddByte($80 OR (wc AND $3F));
    end;
  end; //for
  Result := integer(pch)-integer(@utf8Buf);
end; { WideCharBufToUTF8Buf }

{:Converts UTF-8 encoded buffer into WideChars. Target buffer must be
  pre-allocated and large enough (at most utfByteCount number of WideChars will
  be generated).                                                                 <br>
  RFC 2279 (http://www.ietf.org/rfc/rfc2279.txt) describes the conversion:       <br>
  $00..$7F => $0000..$007F                                                       <br>
  110[bit10..bit6] 10[bit5..bit0] => $0080..$07FF                                <br>
  1110[bit15..bit12] 10[bit11..bit6] 10[bit5..bit0] => $0800..$FFFF
  @param   utf8Buf      UTF-8 encoded buffer.
  @param   utfByteCount Size of utf8Buf, in bytes.
  @param   unicodeBuf   Pre-allocated buffer for WideChars.
  @param   leftUTF8     Number of bytes left in utf8Buf after conversion (0, 1,
                        or 2).
  @returns Number of bytes used in unicodeBuf buffer.
  @since   2.01
}
function UTF8BufToWideCharBuf(const utf8Buf; utfByteCount: integer;
 var unicodeBuf; var leftUTF8: integer): integer;
var
  c1 : byte;
  c2 : byte;
  ch : byte;
  pch: PChar;
  pwc: PWideChar;
begin
  pch := @utf8Buf;
  pwc := @unicodeBuf;
  leftUTF8 := utfByteCount;
  while leftUTF8 > 0 do begin
    ch := byte(pch^);
    Inc(pch);
    if (ch AND $80) = 0 then begin // 1-byte code
      word(pwc^) := ch;
      Inc(pwc);
      Dec(leftUTF8);
    end
    else if (ch AND $E0) = $C0 then begin // 2-byte code
      if leftUTF8 < 2 then
        break;
      c1 := byte(pch^);
      Inc(pch);
      word(pwc^) := (word(ch AND $1F) SHL 6) OR (c1 AND $3F);
      Inc(pwc);
      Dec(leftUTF8,2);
    end
    else begin // 3-byte code
      if leftUTF8 < 3 then
        break;
      c1 := byte(pch^);
      Inc(pch);
      c2 := byte(pch^);
      Inc(pch);
      word(pwc^) :=
        (word(ch AND $0F) SHL 12) OR
        (word(c1 AND $3F) SHL 6) OR
        (c2 AND $3F);
      Inc(pwc);
      Dec(leftUTF8,3);
    end;
  end; //while
  Result := integer(pwc)-integer(@unicodeBuf);
end; { UTF8BufToWideCharBuf }

{:Returns default Ansi codepage for LangID or 1252 in case of error (LangID does
  not specify valid language ID).
  @param   LangID Language ID.
  @returns Default Ansi codepage for LangID or 1252 in case of error.
}
function GetDefaultAnsiCodepage (LangID: LCID): word;
var
  p: array [0..255] of char;
begin
  if GetLocaleInfo (LangID, 4100, p, High (p)) > 0 then
    Result := StrToIntDef(p,1252)
  else
    Result := 1252;
end; { GetDefaultAnsiCodepage }

{ TGpTextFile }

{:Allocates small buffer if not already allocated.
  @since   2.01
}
procedure TGpTextFile.PrepareBuffer;
begin
  if not assigned(tfSmallBuf) then
    GetMem(tfSmallBuf,CTFSSmallBufSize);
end; { TGpTextFile.PrepareBuffer }

{:Allocates buffer for 8/16/8 bit conversions. If requested size is small
  enough, returns pre-allocated buffer, otherwise allocates new buffer.
  @param   size Requested size in bytes.
  @returns Pointer to buffer.
}
function TGpTextFile.AllocTmpBuffer(size: integer): pointer;
begin
  if size <= CTFSSmallBufSize then
    Result := tfSmallBuf
  else
    GetMem(Result,size);
end; { TGpTextFile.AllocTmpBuffer }

{:Frees buffer for 8/16/8 bit conversions. If pre-allocated buffer is passed,
  nothing will be done.
  @param   buffer Conversion buffer.
}
procedure TGpTextFile.FreeTmpBuffer(var buffer: pointer);
begin
  if buffer <> tfSmallBuf then begin
    FreeMem(buffer);
    buffer := nil;
  end;
end; { TGpTextFile.FreeTmpBuffer }

{:Simplest form of Append.
  @param   flags      Create flags. Only cfUse2028 and cfUseLF flags are used.
  @param   bufferSize Size of buffer. 0 means default size (BUF_SIZE, currently
                      64 KB).
  @param   codePage   Code page to be used for 8/16/8 bit conversion. If 0,
                      default code page for currently used language will be
                      used.
  @raises  EGpTextFile if file could not be appended.
}
procedure TGpTextFile.Append(flags: TCreateFlags; bufferSize: integer;
  codePage: word);
begin
  if AppendSafe(flags,bufferSize,0,0,0,codePage) <> hfOK then
    raise EGpTextFile.CreateFmtHelp(sFailedToAppendFile,[FileName],hcTFFailedToAppend);
end; { TGpTextFile.Append }

{:Full form of Append. Will retry if file is locked by another application (if
  diskLockTimeout and diskRetryDelay are specified). Allows caller to specify
  additional options. Does not raise an exception on error (except appending
  reversed Unicode file).
  @param   flags      Create flags. Only cfUse2028 and cfUseLF flags are used.
  @param   bufferSize      Size of buffer. 0 means default size (BUF_SIZE,
                           currently 64 KB).
  @param   diskLockTimeout Max time (in milliseconds) AppendSafe will wait for
                           lock file to become free.
  @param   diskRetryDelay  Delay (in milliseconds) between attempts to open
                           locked file.
  @param   waitObject      Handle of 'terminate' event (semaphore, mutex). If
                           this parameter is specified (not zero) and becomes
                           signalled, AppendSafe will stop trying to open locked
                           file and will exit with.
  @param   codePage        Code page to be used for 8/16/8 bit conversion. If 0,
                           default code page for currently used language will be
                           used.
  @raises  EGpTextFile if file is 'reversed' Unicode file.
}
function TGpTextFile.AppendSafe(flags: TCreateFlags; bufferSize: integer;
  diskLockTimeout, diskRetryDelay: integer; waitObject: THandle;
  codePage: word): THFError;
var
  marker : WideChar;
  marker3: Char;
begin
  try
    if (cfUnicode in flags) and (codePage <> CP_UTF8) then
      codePage := CP_UNICODE;
    PrepareBuffer;
    SetCodepage(codePage);
    Result := ResetEx(1,bufferSize,diskLockTimeout,diskRetryDelay,
      [hfoBuffered,hfoCanCreate],waitObject);
    if Result = hfOK then begin
      tfCFlags := [];
      if FileSize >= SizeOf(WideChar) then begin
        BlockReadUnsafe(marker,SizeOf(WideChar));
        if marker = CUnicodeNormal then begin
          tfCFlags := tfCFlags + [cfUnicode];
          Codepage := CP_UNICODE;
        end
        else if marker = CUnicodeReversed then begin
          tfCFlags := tfCFlags + [cfUnicode,cfReverseByteOrder];
          Codepage := CP_UNICODE;
        end
        else if (marker = CUTF8BOM12) and (FileSize >= 3) then begin
          BlockReadUnsafe(marker3,SizeOf(Char));
          if marker3 = CUTF8BOM3 then begin
            tfCFlags := tfCFlags + [cfUnicode];
            Codepage := CP_UTF8;
          end;
        end;
      end
      else if (FileSize = 0) and IsUnicode then begin
        if Codepage <> CP_UTF8 then
          BlockWriteUnsafe(CUnicodeNormal,SizeOf(WideChar))
        else if cfWriteUTF8BOM in flags then begin
          BlockWriteUnsafe(CUTF8BOM12,SizeOf(WideChar));
          BlockWriteUnsafe(CUTF8BOM3,SizeOf(Char));
        end;
      end;
      if (not IsUnicode) and ((Codepage = CP_UTF8) or (Codepage = CP_UNICODE)) then
        tfCFlags := tfCFlags + [cfUnicode];
      if tfCFlags*[cfUnicode,cfReverseByteOrder] = [cfUnicode,cfReverseByteOrder] then
        raise EGpTextFile.CreateFmtHelp(sCannotAppendReversedUnicodeFile,[FileName],hcTFCannotAppendReversed);
      tfCFlags := tfCFlags + (flags * [cfUse2028,cfUseLF]);
      Seek(FileSize);
    end;
  except
    on EGpTextFile do
      raise;
    on Exception do
      Result := hfError;
  end;
end; { TGpTextFile.Append }

{:Cleanup.
  @since   2.01
}
destructor TGpTextFile.Destroy;
begin
  if assigned(tfSmallBuf) then begin
    FreeMem(tfSmallBuf);
    tfSmallBuf := nil;
  end;
  inherited;
end; { TGpTextFile.Destroy }

{:Checks if file pointer is at end of file.
  @returns True if file pointer is at end of file.
  @raises  EGpHugeFile on Windows errors.
}
function TGpTextFile.EOF: boolean;
begin
  Result := (FilePos >= FileSize);
end; { TGpTextFile.EOF }

{:Checks if file is 16-bit Unicode.
  @returns True if file is 16-bit Unicode.
  @since   2.01
}
function TGpTextFile.Is16bit: boolean;
begin
  Result := IsUnicode and (Codepage <> CP_UTF8);
end; { TGpTextFile.Is16bit }

{:Checks if file is Unicode (UCS-2 or UTF-8 encoding).
  @returns True if file is Unicode.
}
function TGpTextFile.IsUnicode: boolean;
begin
  Result := (cfUnicode in tfCFlags);
end; { TGpTextFile.IsUnicode }

{:Reads line from file. If file is 8-bit, LF, CR, CRLF, and LFCR are considered
  end-of-line terminators (if included in AcceptedDelimiters). If file is
  16-bit, both /000D/000A/ and /2028/ are considered end-of-line terminators (if
  included in AcceptedDelimiters). If file is 8-bit, line is converted to
  Unicode according to code page specified in Append, Reset or Rewrite.
  @returns Line without terminator characters.
  @raises  EGpHugeFile on Windows errors.
  @seeAlso Append, Reset, Rewrite
}
function TGpTextFile.Readln: WideString;
var
  buf     : array [1..1028] of byte; // size must be even for Unicode
  bufPos  : HugeInt;
  delimLen: integer;
  delimPos: integer;
  leftUtf8: integer;
  leof    : boolean;
  resLn   : WideString;
  size    : cardinal;
  uniBytes: integer;
  utf8Ln  : string;

  procedure ReverseBlock;
  var
    tmp: byte;
    i  : cardinal;
  begin
  {$R-}
    i := 1;
    while i < size do begin
      tmp := buf[i];
      buf[i] := buf[i+1];
      buf[i+1] := tmp;
      Inc(i,2);
    end; //while
  {$R+}
  end; { ReverseBlock }

  procedure FetchBlock;
  begin
    bufPos := FilePos;
    BlockRead(buf,SizeOf(buf)-2,size);
    if cfReverseByteOrder in tfCFlags then
      ReverseBlock;
    leof := (size < (SizeOf(buf)-2));
    // simplify LocateDelimiter
    if not leof then
      Dec(size,2)
    else begin
      buf[size+1] := 0;
      buf[size+2] := 0;
    end;
  end; { FetchBlock }

  procedure LocateDelimiter;
  var
    i  : integer;
    isu: boolean;
  begin
  {$R-}
    delimPos := size+1;
    delimLen := 0;
    isu := IsUnicode and (Codepage <> CP_UTF8);
    for i := 1 to size do begin
      if isu then begin
        if ((AcceptedDelimiters = []) or
            (ld000D000A in AcceptedDelimiters)) and
           ((buf[i] = $0D) and (buf[i+1] = $00) and
            (buf[i+2] = $0A) and (buf[i+3] = $00)) then
        begin
          delimPos := i;
          delimLen := 4;
          break;
        end
        else if ((AcceptedDelimiters = []) or
                 (ld2028 in AcceptedDelimiters)) and
                ((buf[i] = $28) and (buf[i+1] = $20)) then
        begin
          delimPos := i;
          delimLen := 2;
          break;
        end;
      end
      else begin
        if ((AcceptedDelimiters = []) or
            (ldCRLF in AcceptedDelimiters)) and
           ((buf[i] = $0D) and (buf[i+1] = $0A)) then
        begin
          delimPos := i;
          delimLen := 2;
          break;
        end
        else if ((AcceptedDelimiters = []) or
                 (ldLFCR in AcceptedDelimiters)) and
                 ((buf[i] = $0A) and (buf[i+1] = $0D)) then
        begin
          delimPos := i;
          delimLen := 2;
          break;
        end
        else if (((AcceptedDelimiters = []) or (ldCR in AcceptedDelimiters)) and (buf[i] = $0D)) or
                (((AcceptedDelimiters = []) or (ldLF in AcceptedDelimiters)) and (buf[i] = $0A)) then
        begin
          delimPos := i;
          delimLen := 1;
          break;
        end;
      end;
    end; //for
  {$R+}
  end; { LocateDelimiter }

  procedure PasteAndSeek;  
  var
    ansiLn: AnsiString;
    wideLn: WideString;
  begin
    if IsUnicode then begin
      if Codepage = CP_UTF8 then begin
        SetLength(ansiLn,delimPos-1);
        if Length(ansiLn) > 0 then
          Move(buf[1],ansiLn[1],Length(ansiLn));
        utf8Ln := utf8Ln + ansiLn;
      end
      else begin
        SetLength(wideLn,(delimPos-1) div SizeOf(WideChar));
        if Length(wideLn) > 0 then
          Move(buf[1],wideLn[1],Length(wideLn)*SizeOf(WideChar));
        resLn := resLn + wideLn;
      end;
    end
    else begin
      SetLength(ansiLn,delimPos-1);
      if Length(ansiLn) > 0 then
        Move(buf[1],ansiLn[1],Length(ansiLn));
      resLn := resLn + StringToWideString(ansiLn,tfCodePage);
    end;
    Seek(bufPos+delimPos+delimLen-1);
  end; { PasteAndSeek }

begin { TGpTextFile.Readln }
  try
    if Codepage = CP_UTF8 then
      utf8Ln := ''
    else
      resLn := '';
    repeat
      FetchBlock;
      if size = 0 then
        break;
      LocateDelimiter;
      PasteAndSeek;
      if delimLen > 0 then
        break;
    until leof;
    if Codepage = CP_UTF8 then begin
      if utf8Ln = '' then
        Result := ''
      else begin
        SetLength(Result,Length(utf8Ln)); // worst case
        uniBytes := UTF8BufToWideCharBuf(utf8Ln[1],Length(utf8Ln),Result[1],leftUtf8);
        SetLength(Result,uniBytes div SizeOf(WideChar));
      end;
    end
    else
      Result := resLn;
  except
    on E: EGpTextFile do raise;
    on E: EGpHugeFile do raise;
    on E: Exception   do raise EGpTextFile.CreateHelp(E.Message,hcTFUnexpected);
  end;
end; { TGpTextFile.Readln }

{:Simplest form of Reset.
  @param   bufferSize Size of buffer. 0 means default size (BUF_SIZE, currently
                      64 KB).
  @param   flags      Open flags. 
  @param   codePage   Code page to be used for 8/16/8 bit conversion. If 0,
                      default code page for currently used language will be
                      used.
  @raises  EGpTextFile if file could not be reset.
}
procedure TGpTextFile.Reset(flags: TOpenFlags; bufferSize: integer; 
  codePage: word);
begin
  if ResetSafe(flags,bufferSize,0,0,0,codePage) <> hfOK then
    raise EGpTextFile.CreateFmtHelp(sFailedToResetFile,[FileName],hcTFFailedToReset);
end; { TGpTextFile.Reset }

{:Full form of Reset. Will retry if file is locked by another application (if
  diskLockTimeout and diskRetryDelay are specified). Allows caller to specify
  additional options. Does not raise an exception on error.
  @param   flags           Open flags.
  @param   bufferSize      Size of buffer. 0 means default size (BUF_SIZE,
                           currently 64 KB).
  @param   diskLockTimeout Max time (in milliseconds) Reset will wait for lock
                           file to become free.
  @param   diskRetryDelay  Delay (in milliseconds) between attempts to open
                           locked file.
  @param   waitObject      Handle of 'terminate' event (semaphore, mutex). If
                           this parameter is specified (not zero) and becomes
                           signalled, Reset will stop trying to open locked file
                           and will exit with.
  @param   codePage        Code page to be used for 8/16/8 bit conversion. If 0,
                           default code page for currently used language will be
                           used.
  @raises  EGpHugeFile on Windows errors.
}
function TGpTextFile.ResetSafe(flags: TOpenFlags; bufferSize: integer;
  diskLockTimeout, diskRetryDelay: integer; waitObject: THandle;
  codePage: word): THFError;
var
  marker : WideChar;
  marker3: Char;
  options: THFOpenOptions;
begin
  try
    SetCodepage(codePage);
    PrepareBuffer;
    options := [hfoBuffered];
    if ofCloseOnEOF in flags then
      options := options + [hfoCloseOnEOF];
    Result := ResetEx(1,bufferSize,diskLockTimeout,diskRetryDelay,options,waitObject);
    if Result = hfOK then begin
      tfCFlags := [];
      if FileSize >= SizeOf(WideChar) then begin
        BlockReadUnsafe(marker,SizeOf(WideChar));
        if marker = CUnicodeNormal then begin
          tfCFlags := tfCFlags + [cfUnicode];
          Codepage := 1200;
        end
        else if marker = CUnicodeReversed then begin
          tfCFlags := tfCFlags + [cfUnicode,cfReverseByteOrder];
          Codepage := 1200;
        end
        else if (marker = CUTF8BOM12) and (FileSize >= 3) then begin
          BlockReadUnsafe(marker3,SizeOf(Char));
          if marker3 = CUTF8BOM3 then begin
            tfCFlags := tfCFlags + [cfUnicode];
            Codepage := CP_UTF8;
          end;
        end;
        if not IsUnicode then
          Seek(0);
      end;
      if (not IsUnicode) and ((Codepage = CP_UTF8) or (Codepage = CP_UNICODE)) then
        tfCFlags := tfCFlags + [cfUnicode];
    end;
  except
    Result := hfError;
  end;
end; { TGpTextFile.ResetSafe }

{:Simplest form of Rewrite.
  @param   flags      Create flags. 
  @param   bufferSize Size of buffer. 0 means default size (BUF_SIZE, currently
                      64 KB).
  @param   codePage   Code page to be used for 8/16/8 bit conversion. If 0,
                      default code page for currently used language will be
                      used.
  @raises  EGpTextFile if file could not be appended.
}
procedure TGpTextFile.Rewrite(flags: TCreateFlags; bufferSize: integer;
  codePage: word);
begin
  if RewriteSafe(flags,bufferSize,0,0,0,codePage) <> hfOK then
    raise EGpTextFile.CreateFmtHelp(sFailedToRewriteFile,[FileName],hcTFFailedToRewrite);
end; { TGpTextFile.Rewrite }

{:Full form of Rewrite. Will retry if file is locked by another application (if
  diskLockTimeout and diskRetryDelay are specified). Allows caller to specify
  additional options. Does not raise an exception on error.
  @param   flags           Create flags.
  @param   bufferSize      Size of buffer. 0 means default size (BUF_SIZE,
                           currently 64 KB).
  @param   diskLockTimeout Max time (in milliseconds) Rewrite will wait for
                           lock file to become free.
  @param   diskRetryDelay  Delay (in milliseconds) between attempts to open
                           locked file.
  @param   waitObject      Handle of 'terminate' event (semaphore, mutex). If
                           this parameter is specified (not zero) and becomes
                           signalled, Rewrite will stop trying to open locked
                           file and will exit with.
  @param   codePage        Code page to be used for 8/16/8 bit conversion. If 0,
                           default code page for currently used language will be
                           used.
  @raises  EGpTextFile if file is 'reversed' Unicode file.
  @raises  EGpHugeFile on Windows errors.
}
function TGpTextFile.RewriteSafe(flags: TCreateFlags; bufferSize: integer;
  diskLockTimeout, diskRetryDelay: integer; waitObject: THandle;
  codePage: word): THFError;
begin
  if (cfUnicode in flags) and (codePage <> CP_UTF8) then
    codePage := CP_UNICODE;
  PrepareBuffer;
  if ((Codepage = CP_UTF8) or (Codepage = CP_UNICODE)) then
    flags := flags + [cfUnicode];    
  if flags*[cfUnicode,cfReverseByteOrder] = [cfUnicode,cfReverseByteOrder] then
    raise EGpTextFile.CreateFmtHelp(sCannotWriteReversedUnicodeFile,[FileName],hcTFCannotWriteReversed);
  try
    SetCodepage(codePage);
    Result := RewriteEx(1,bufferSize,diskLockTimeout,diskRetryDelay,[hfoBuffered],waitObject);
    if Result = hfOK then begin
      Truncate;
      tfCFlags := flags;
      if IsUnicode then begin
        if Codepage <> CP_UTF8 then
          BlockWriteUnsafe(CUnicodeNormal,SizeOf(WideChar))
        else if cfWriteUTF8BOM in flags then begin
          BlockWriteUnsafe(CUTF8BOM12,SizeOf(WideChar));
          BlockWriteUnsafe(CUTF8BOM3,SizeOf(Char));
        end;
      end;
    end;
  except
    Result := hfError;
  end;
end; { TGpTextFile.RewriteSafe }

{:Internal method that sets current code page or locates default code page if
  0 is passed as a parameter.
  @param   cp Code page number or 0 for default code page.
}
procedure TGpTextFile.SetCodepage(cp: word);
begin
  if (cp = CP_UTF8) or (cp = CP_UNICODE) then begin
    tfCodePage := cp;
    tfCFlags := tfCFlags + [cfUnicode];
  end
  else begin
    if (cp = 0) and (not IsUnicode) then
      tfCodePage := GetDefaultAnsiCodepage(GetKeyboardLayout(GetCurrentThreadId) and $FFFF)
    else
      tfCodePage := cp;
    if not ((tfCodePage = 0) or (tfCodePage = CP_UNICODE)) then
      tfCFlags := tfCFlags - [cfUnicode];
  end;
end; { TGpTextFile.SetCodepage }

{:Writes string to the text file. If file is 8-bit, string is converted
  according to Codepage property.
  @param   ws String to be written.
  @raises  EGpHugeFile on Windows errors.
}
procedure TGpTextFile.WriteString(ws: WideString);
var
  ansiLn  : AnsiString;
  numBytes: integer;
  numChar : integer;
  tmpBuf  : pointer;
begin
  if ws = '' then Exit;
  if IsUnicode then begin
    if Codepage = CP_UTF8 then begin
      numChar := Length(ws);
      tmpBuf := AllocTmpBuffer(numChar*3); // worst case - 3 bytes per character
      try
        numBytes := WideCharBufToUTF8Buf(ws[1],Length(ws)*SizeOf(WideChar),tmpBuf^);
        BlockWriteUnsafe(tmpBuf^,numBytes);
      finally FreeTmpBuffer(tmpBuf); end;
    end
    else
      BlockWriteUnsafe(ws[1],Length(ws)*SizeOf(WideChar))
  end
  else begin
    ansiLn := WideStringToString(ws,tfCodePage);
    BlockWriteUnsafe(ansiLn[1],Length(ansiLn));
  end;
end; { TGpTextFile.WriteString }

{:Writes array of values to the text file. If file is 8-bit, values are
  converted according to Codepage property.
  @param   Values.
  @raises  EGpTextFile on unsupported parameter.
  @raises  EGpHugeFile on Windows errors.
}
procedure TGpTextFile.Write(params: array of const);
var
  i     : integer;
  wideLn: WideString;
const
  BoolChars: array [boolean] of char = ('F','T');
begin
  try
    wideLn := '';
    for i := 0 to High(params) do begin
      with params[i] do begin
        case VType of
          vtInteger:    wideLn := wideLn + IntToStr(VInteger);
          vtBoolean:    wideLn := wideLn + BoolChars[VBoolean];
          vtChar:       wideLn := wideLn + StringToWideString(VChar,tfCodePage);
          vtExtended:   wideLn := wideLn + StringToWideString(FloatToStr(VExtended^),tfCodePage);
          vtString:     wideLn := wideLn + StringToWideString(VString^,tfCodePage);
          vtPointer:    wideLn := wideLn + IntToHex(integer(VPointer),8);
          vtPChar:      wideLn := wideLn + StringToWideString(VPChar,tfCodePage);
          vtObject:     wideLn := wideLn + StringToWideString(VObject.ClassName,tfCodePage);
          vtClass:      wideLn := wideLn + StringToWideString(VClass.ClassName,tfCodePage);
          vtWideChar:   wideLn := wideLn + VWideChar;
          vtPWideChar:  wideLn := wideLn + VPWideChar^;
          vtAnsiString: wideLn := wideLn + StringToWideString(string(VAnsiString),tfCodePage);
          vtCurrency:   wideLn := wideLn + StringToWideString(CurrToStr(VCurrency^),tfCodePage);
          vtVariant:    wideLn := wideLn + StringToWideString(string(VVariant^),tfCodePage);
          vtWideString: wideLn := wideLn + WideString(VWideString);
          vtInt64:      wideLn := wideLn + IntToStr(VInt64^);
          else raise EGpTextFile.CreateFmtHelp(sInvalidParameter,[FileName],hcTFInvalidParameter);
        end;
      end;
    end;
    WriteString(wideLn);
  except
    on E: EGpTextFile do raise;
    on E: EGpHugeFile do raise;
    on E: Exception   do raise EGpTextFile.CreateHelp(E.Message,hcTFUnexpected);
  end;
end; { TGpTextFile.Write }

{:Writes line to the text file. If file is 8-bit, values are converted
  according to Codepage property. Uses line delimiter set in Rewrite/Append.
  @param   ln Line to be written.
  @raises  EGpHugeFile on Windows errors.
  @seeAlso Rewrite, Append
}
procedure TGpTextFile.Writeln(ln: WideString);
var
  ansiLn: AnsiString;
begin
  try
    WriteString(ln);
    if IsUnicode then begin
      if cfUse2028 in tfCFlags then 
        WriteString(WideChar($2028))
      else begin
        WriteString(WideChar($000D));
        WriteString(WideChar($000A));
      end;
    end
    else begin
      if cfUseLF in tfCFlags then
        ansiLn := Char($0D)
      else
        ansiLn := Char($0D)+Char($0A);
      BlockWriteUnsafe(ansiLn[1],Length(ansiLn));
    end;
  except
    on E: EGpTextFile do raise;
    on E: EGpHugeFile do raise;
    on E: Exception   do raise EGpTextFile.CreateHelp(E.Message,hcTFUnexpected);
  end;
end; { TGpTextFile.Writeln }

{:Writes array of values to the text file then terminates the line with line
  delimiter. If file is 8-bit, values are converted according to Codepage
  property. Uses line delimiter set in Rewrite/Append.
  @param   Values.
  @raises  EGpTextFile on unsupported parameter.
  @raises  EGpHugeFile on Windows errors.
  @seeAlso Rewrite, Append
}
procedure TGpTextFile.Writeln(params: array of const);
begin
  Write(params);
  Writeln('');
end; { TGpTextFile.Writeln }

{ TGpTextFileStream }

{:Allocates buffer for 8/16/8 bit conversions. If requested size is small
  enough, returns pre-allocated buffer, otherwise allocates new buffer.
  @param   size Requested size in bytes.
  @returns Pointer to buffer.
}
function TGpTextFileStream.AllocBuffer(size: integer): pointer;
begin
  if size <= CTFSSmallBufSize then
    Result := tfsSmallBuf
  else
    GetMem(Result,size);
end; { TGpTextFileStream.AllocBuffer }

{:Initializes stream and opens file in required access mode.
  @param   fileName    Name of file to be accessed.
  @param   access      Required access mode.
  @param   openFlags   Open flags (used when access mode is accReset).
  @param   createFlags Create flags (used when access mode is accRewrite or
                       accAppend).
  @param   codePage    Code page to be used for 8/16/8 bit conversions. If set
                       to 0, current default code page will be used.
}
constructor TGpTextFileStream.Create(const fileName: string;
  access: TGpHugeFileStreamAccess; openFlags: TOpenFlags;
  createFlags: TCreateFlags; codePage: word);
var
  openOptions: THFOpenOptions;
begin
  inherited Create;
  if (cfUnicode in createFlags) and (codePage <> CP_UTF8) then
    codePage := CP_UNICODE;
  openOptions := [hfoBuffered];
  if (access = accRead) and (ofCloseOnEOF in openFlags) then
    openOptions := openOptions + [hfoCloseOnEOF];
  tfsStream := TGpHugeFileStream.Create(fileName,access,openOptions);
  tfsExternalHFS := false;
  tfsAccess := access;
  tfsOpenFlags := openFlags;
  tfsCreateFlags := createFlags;
  SetCodepage(codePage);
  GetMem(tfsSmallBuf,CTFSSmallBufSize);
  PrepareStream;
end; { TGpTextFileStream.Create }

{:Initializes stream and assigns to it already open TGpHugeFileStream object.
  @param   hfs         TGpHugeFileStream to be used for data storage.
  @param   access      Required access mode.
  @param   openFlags   Open flags (used when access mode is accReset).
  @param   createFlags Create flags (used when access mode is accRewrite or
                       accAppend).
  @param   codePage    Code page to be used for 8/16/8 bit conversions. If set
                       to 0, current default code page will be used.
}
constructor TGpTextFileStream.CreateFromHandle(
  hfs: TGpHugeFileStream; access: TGpHugeFileStreamAccess;
  openFlags: TOpenFlags; createFlags: TCreateFlags; codePage: word);
begin
  inherited Create;
  if (cfUnicode in createFlags) and (codePage <> CP_UTF8) then
    codePage := CP_UNICODE;
  tfsExternalHFS := true;
  tfsStream := hfs;
  tfsAccess := access;
  tfsOpenFlags := openFlags;
  tfsCreateFlags := createFlags;
  SetCodepage(codePage);
  GetMem(tfsSmallBuf,CTFSSmallBufSize);
  PrepareStream;
end; { TGpTextFileStream.CreateFromHandle }

{:Cleanup. Destroys data stream only if it was created in constructor.
}
destructor TGpTextFileStream.Destroy;
begin
  if (not tfsExternalHFS) and assigned(tfsStream) then begin
    tfsStream.Free;
    tfsStream := nil;
  end;
  FreeMem(tfsSmallBuf);
  tfsReadlnBuf.Free;
  tfsReadlnBuf := nil;
  inherited Destroy;
end; { TGpTextFileStream.Destroy }

{:Frees buffer for 8/16/8 bit conversions. If pre-allocated buffer is passed,
  nothing will be done.
  @param   buffer Conversion buffer.
}
procedure TGpTextFileStream.FreeBuffer(var buffer: pointer);
begin
  if buffer <> tfsSmallBuf then begin
    FreeMem(buffer);
    buffer := nil;
  end;
end; { TGpTextFileStream.FreeBuffer }

{:Returns file name.
  @returns Returns file name or empty string if file is not open.
}
function TGpTextFileStream.GetFileName: string;
begin
  if assigned(tfsStream) then
    Result := tfsStream.FileName
  else
    Result := '';
end; { TGpTextFileStream.GetFileName }

{:Returns file size. Better compatibility with hfCloseOnEOF files than default
  TStream.GetSize.
  @returns Returns file size in bytes or -1 if file is not open.
}
function TGpTextFileStream.GetSize: longint;
begin
  if assigned(tfsStream) then
    Result := tfsStream.Size
  else
    Result := -1;
end; { TGpTextFileStream.GetSize }

{:Returns last Windows error code.
  @returns Last Windows error code.
}
function TGpTextFileStream.GetWindowsError: DWORD;
begin
  if tfsWindowsError <> 0 then
    Result := tfsWindowsError
  else if assigned(tfsStream) then
    Result := tfsStream.WindowsError
  else
    Result := 0;
end; { TGpTextFileStream.GetWindowsError }

{:Checks if file is 16-bit Unicode.
  @returns True if file is 16-bit Unicode.
  @since   2.01
}
function TGpTextFileStream.Is16bit: boolean;
begin
  Result := IsUnicode and (Codepage <> CP_UTF8);
end; { TGpTextFileStream.Is16bit }

{:Checks if stream is Unicode (UCS-2 or UTF-8 encoding).
  @returns True if stream is Unicode.
}
function TGpTextFileStream.IsUnicode: boolean;
begin
  Result := (cfUnicode in tfsCreateFlags);
end; { TGpTextFileStream.IsUnicode }

{:Prepares stream for read or write operation.
  @raises EGpTextFileStream if caller tries to rewrite or append 'reverse'
          Unicode stream.
}
procedure TGpTextFileStream.PrepareStream;
var
  marker : WideChar;
  marker3: Char;
begin
  case tfsAccess of
    accRead:
      begin
        tfsCreateFlags := [];
        if tfsStream.Size >= SizeOf(WideChar) then begin
          tfsStream.Read(marker,SizeOf(WideChar));
          if marker = CUnicodeNormal then begin
            tfsCreateFlags := tfsCreateFlags + [cfUnicode];
            Codepage := CP_UNICODE;
          end
          else if marker = CUnicodeReversed then begin
            tfsCreateFlags := tfsCreateFlags + [cfUnicode,cfReverseByteOrder];
            Codepage := CP_UNICODE;
          end
          else if (marker = CUTF8BOM12) and (tfsStream.Size >= 3) then begin
            tfsStream.Read(marker3,SizeOf(Char));
            if marker3 = CUTF8BOM3 then begin
              tfsCreateFlags := tfsCreateFlags + [cfUnicode];
              Codepage := CP_UTF8;
            end;
          end;
          if not IsUnicode then
            tfsStream.Position := 0;
        end;
        if (not IsUnicode) and ((Codepage = CP_UTF8) or (Codepage = CP_UNICODE)) then
          tfsCreateFlags := [cfUnicode];
      end; //accRead
    accWrite:
      begin
        if ((Codepage = CP_UTF8) or (Codepage = CP_UNICODE)) then
          tfsCreateFlags := tfsCreateFlags + [cfUnicode];
        if tfsCreateFlags*[cfUnicode,cfReverseByteOrder] = [cfUnicode,cfReverseByteOrder] then
          raise EGpTextFileStream.CreateFmtHelp(sCannotWriteReversedUnicodeStream,[FileName],hcTFCannotWriteReversed);
        tfsStream.Size := 0;
        if IsUnicode then begin
          if Codepage <> CP_UTF8 then
            tfsStream.Write(CUnicodeNormal,SizeOf(WideChar))
          else if cfWriteUTF8BOM in tfsCreateFlags then begin
            tfsStream.Write(CUTF8BOM12,SizeOf(WideChar));
            tfsStream.Write(CUTF8BOM3,SizeOf(Char));
          end;
        end;
      end; //accWrite
    accReadWrite:
      begin
        if ((Codepage = CP_UTF8) or (Codepage = CP_UNICODE)) then
          tfsCreateFlags := tfsCreateFlags + [cfUnicode];
        if tfsCreateFlags*[cfUnicode,cfReverseByteOrder] = [cfUnicode,cfReverseByteOrder] then
          raise EGpTextFileStream.CreateFmtHelp(sCannotAppendReversedUnicodeStream,[FileName],hcTFCannotAppendReversed);
        if (tfsStream.Size = 0) and IsUnicode then begin
          if Codepage <> CP_UTF8 then
            tfsStream.Write(CUnicodeNormal,SizeOf(WideChar))
          else if cfWriteUTF8BOM in tfsCreateFlags then begin
            tfsStream.Write(CUTF8BOM12,SizeOf(WideChar));
            tfsStream.Write(CUTF8BOM3,SizeOf(Char));
          end;
        end;
      end; //accReadWrite
    accAppend:
      begin
        tfsCreateFlags := [];
        if tfsStream.Size >= SizeOf(WideChar) then begin
          tfsStream.Position := 0;
          tfsStream.Read(marker,SizeOf(WideChar));
          if marker = CUnicodeNormal then begin
            tfsCreateFlags := tfsCreateFlags + [cfUnicode];
            Codepage := CP_UNICODE;
          end
          else if marker = CUnicodeReversed then begin
            tfsCreateFlags := tfsCreateFlags + [cfUnicode,cfReverseByteOrder];
            Codepage := CP_UNICODE;
          end
          else if (marker = CUTF8BOM12) and (tfsStream.Size >= 3) then begin
            tfsStream.Read(marker3,SizeOf(Char));
            if marker3 = CUTF8BOM3 then begin
              tfsCreateFlags := tfsCreateFlags + [cfUnicode];
              Codepage := CP_UTF8;
            end;
          end;
          tfsStream.Position := tfsStream.Size;
        end
        else if (tfsStream.Size = 0) and IsUnicode then begin
          if Codepage <> CP_UTF8 then
            tfsStream.Write(CUnicodeNormal,SizeOf(WideChar))
          else if cfWriteUTF8BOM in tfsCreateFlags then begin
            tfsStream.Write(CUTF8BOM12,SizeOf(WideChar));
            tfsStream.Write(CUTF8BOM3,SizeOf(Char));
          end;
        end;
        if (not IsUnicode) and ((Codepage = CP_UTF8) or (Codepage = CP_UNICODE)) then
          tfsCreateFlags := tfsCreateFlags + [cfUnicode];
        if tfsCreateFlags*[cfUnicode,cfReverseByteOrder] = [cfUnicode,cfReverseByteOrder] then
          raise EGpTextFileStream.CreateFmtHelp(sCannotAppendReversedUnicodeStream,[FileName],hcTFCannotAppendReversed);
      end; //accAppend
  end; //case
end; { TGpTextFileStream.PrepareStream }

{:Reads 'count' number of bytes from stream. 'Count' must be an even number as
  data is always returned in Unicode format (two bytes per character). If file
  is 8-bit, data is converted to Unicode according to code page specified in
  constructor.
  @param   buffer Buffer for read data.
  @param   count  Number of bytes to be read.
  @returns Number of bytes actually read.
  @raises  EGpTextFileStream if 'count' is odd.
  @raises  EGpTextFileStream if conversion from 8-bit to Unicode failes.
  @raises  EGpHugeFile on Windows errors.
}
function TGpTextFileStream.Read(var buffer; count: longint): longint;
var
  bufPtr   : pointer;
  bytesConv: integer;
  bytesLeft: integer;
  bytesRead: integer;
  numChar  : integer;
  tmpBuf   : pointer;
begin
  if IsUnicode then begin
    if Codepage = CP_UTF8 then begin
      numChar := count div SizeOf(WideChar);
      tmpBuf := AllocBuffer(numChar);
      try
        bufPtr := @buffer;
        Result := 0;
        bytesLeft := 0;
        repeat
          // at least numChar UTF-8 bytes are needed for numChar WideChars
          bytesRead := tfsStream.Read(pointer(integer(tmpBuf)+bytesLeft)^,numChar);
          bytesConv := UTF8BufToWideCharBuf(tmpBuf^,bytesRead+bytesLeft,bufPtr^,bytesLeft);
          Result := Result + bytesConv;
          if bytesRead <> numChar then // end of stream
            break;
          numChar := numChar - (bytesConv div SizeOf(WideChar));
          bufPtr := pointer(integer(bufPtr) + bytesConv);
          if bytesLeft > 0 then
            Move(pointer(integer(tmpBuf)+bytesRead-bytesLeft)^,tmpBuf^,bytesLeft);
        until numChar = 0;
      finally FreeBuffer(tmpBuf); end;
    end
    else
      Result := tfsStream.Read(buffer,count);
  end
  else begin
    if Odd(count) then
      raise EGpTextFileStream.CreateFmtHelp(sCannotConvertOddNumberOfBytes,[FileName,count],hcTFCannotConvertOdd)
    else begin
      numChar := count div SizeOf(WideChar);
      tmpBuf := AllocBuffer(numChar);
      try
        bytesRead := tfsStream.Read(tmpBuf^,numChar);
        numChar := MultiByteToWideChar(tfsCodePage, MB_PRECOMPOSED,
          PChar(tmpBuf), bytesRead, PWideChar(@buffer), numChar);
        Win32Check(numChar <> 0,'Read');
        Result := numChar * SizeOf(WideChar);
      finally FreeBuffer(tmpBuf); end;
    end;
  end;
end; { TGpTextFileStream.Read }

{:Reads one text line stream. If file is 8-bit, LF, CR, CRLF, and LFCR are
  considered end-of-line terminators (if included in AcceptedDelimiters). If
  file is 16-bit, both /000D/000A/ and /2028/ are considered end-of-line
  terminators (if included in AcceptedDelimiters). If file is 8-bit, line is
  converted to Unicode according to code page specified in constructor.
  <b>This function is quite slow.</b>
  @returns One line of text.
  @raises  EGpTextFileStream if conversion from 8-bit to Unicode failes.
  @raises  EGpHugeFile on Windows errors.
}
function TGpTextFileStream.Readln: WideString;
var
  lastCh  : WideChar;
  numCh   : integer;
  wch     : WideChar;

  function Reverse(w: word): word;
  begin
    if cfReverseByteOrder in tfsCreateFlags then begin
      WordRec(Result).Hi := WordRec(w).Lo;
      WordRec(Result).Lo := WordRec(w).Hi;
    end
    else
      Result := w;
  end; { Readln }

  procedure ReverseResult;
  var
    ich: integer;
    pwc: PWord;
  begin
    if cfReverseByteOrder in tfsCreateFlags then begin
      pwc := @Result[1];
      for ich := 1 to Length(Result) div SizeOf(WideChar) do begin
        WordRec(pwc^).Hi := WordRec(pwc^).Lo;
        WordRec(pwc^).Lo := WordRec(pwc^).Hi;
        Inc(pwc);
      end; //for
    end;
  end; { ReverseBlock }

begin { TGpTextFileStream.Readln }
  if assigned(tfsReadlnBuf) then
    tfsReadlnBuf.Clear
  else
    tfsReadlnBuf := TMemoryStream.Create;
  lastCh := #0;
  numCh := 0;
  repeat
    if Read(wch,SizeOf(WideChar)) <> SizeOf(WideChar) then
      break; // EOF
    if (((AcceptedDelimiters = []) or ([ldLF, ldCRLF]*AcceptedDelimiters <> [])) or
        (IsUnicode and ((AcceptedDelimiters = []) or (ld000D000A in AcceptedDelimiters)))) and
       (wch = WideChar(Reverse($000A))) then begin
      if (((AcceptedDelimiters = []) or ([ldLFCR]*AcceptedDelimiters <> [])) or
          (IsUnicode and ((AcceptedDelimiters = []) or (ld000D000A in AcceptedDelimiters)))) and
         (lastCh = WideChar(Reverse($000D))) then
        numCh := 1;
      break;
    end
    else if (([ldCR, ldLFCR]*AcceptedDelimiters <> [])) and
             (wch = WideChar(Reverse($000D))) then begin
      if (([ldLFCR]*AcceptedDelimiters <> [])) and
          (lastCh = WideChar(Reverse($000A))) then
        numCh := 1;
      break;
    end
    else if IsUnicode and
            ((AcceptedDelimiters = []) or (ld2028 in AcceptedDelimiters)) and
            (wch = WideChar(Reverse($2028))) then
      break;
    tfsReadlnBuf.Write(wch,SizeOf(WideChar));
    lastCh := wch;
  until false;
  SetLength(Result,(tfsReadlnBuf.Size-numCh*SizeOf(WideChar)) div SizeOf(WideChar));
  if Result <> '' then
    Move(tfsReadlnBuf.Memory^,Result[1],tfsReadlnBuf.Size-numCh*SizeOf(WideChar));
  ReverseResult;
end; { TGpTextFileStream.Readln }

{:Repositions stream pointer.
  @param   offset Offset from start, current position, or end of stream (as set
                  by the 'mode' parameter) in bytes.
  @param   mode   Specifies starting point for offset calculation
                  (soFromBeginning, soFromCurrent, soFromEnd).
  @returns New position of stream pointer.
  @raises  EGpHugeFile on Windows errors.
  @raises  EGpHugeFileStream on invalid value of 'mode' parameter.
}
function TGpTextFileStream.Seek(offset: longint; mode: word): longint;
begin
  Result := tfsStream.Seek(offset,mode);
end; { TGpTextFileStream.Seek }

{:Internal method that sets current code page or locates default code page if
  0 is passed as a parameter.
  @param   cp Code page number or 0 for default code page.
}
procedure TGpTextFileStream.SetCodepage(cp: word);
begin
  if (cp = CP_UTF8) or (cp = CP_UNICODE) then begin
    tfsCodePage := cp;
    tfsCreateFlags := tfsCreateFlags + [cfUnicode];
  end
  else begin
    if (cp = 0) and (not IsUnicode) then
      tfsCodePage := GetDefaultAnsiCodepage(GetKeyboardLayout(GetCurrentThreadId) and $FFFF)
    else
      tfsCodePage := cp;
    if not ((tfsCodePage = 0) or (tfsCodePage = CP_UNICODE)) then
      tfsCreateFlags := tfsCreateFlags - [cfUnicode];
  end;
end; { TGpTextFileStream.SetCodepage }

{:Sets stream size. Truncates underlying file at specified position.
  @param   newSize New stream size.
  @raises  EGpHugeFile on Windows errors.
}
procedure TGpTextFileStream.SetSize(newSize: longint);
begin
  tfsStream.Size := newSize;
end; { TGpTextFileStream.SetSize }

procedure TGpTextFileStream.Win32Check(condition: boolean; method: string);
{:Checks condition and creates appropriately formatted EGpTextFileStream
  exception.
  @param   condition If false, Win32Check will generate an exception.
  @param   method    Name of TGpTextFileStream method that called Win32Check.
  @raises  EGpTextFileStream if (not condition).
}
var
  Error: EGpTextFileStream;
begin
  if not condition then begin
    tfsWindowsError := GetLastError;
    if tfsWindowsError <> ERROR_SUCCESS then
      Error := EGpTextFileStream.CreateFmtHelp(sStreamFailed+SWin32Error,
        [method,FileName,tfsWindowsError,SysErrorMessage(tfsWindowsError)],
        hcTFWindowsError)
    else
      Error := EGpTextFileStream.CreateFmtHelp(sStreamFailed+SUnkWin32Error,
        [method,FileName],hcTFUnknownWindowsError);
    raise Error;
  end;
end; { TGpTextFileStream.Win32Check }

{:Writes 'count' number of bytes to stream. 'Count' must be an even number as
  data is always expected in Unicode format (two bytes per character). If file
  is 8-bit, data is converted from Unicode according to code page specified in
  constructor.
  @param   buffer Data to be written.
  @param   count  Number of bytes to be written.
  @returns Number of bytes actually written.
  @raises  EGpTextFileStream if 'count' is odd.
  @raises  EGpTextFileStream if conversion from 8-bit to Unicode failes.
  @raises  EGpHugeFile on Windows errors.
}
function TGpTextFileStream.Write(const buffer; count: longint): longint;
var
  leftUTF8  : integer;
  numBytes  : integer;
  numChar   : integer;
  tmpBuf    : pointer;
  uniBuf    : pointer;
  utfWritten: integer;
begin
  if IsUnicode then begin
    if Codepage = CP_UTF8 then begin
      numChar := count div SizeOf(WideChar);
      tmpBuf := AllocBuffer(numChar*3); // worst case - 3 bytes per character
      try
        numBytes := WideCharBufToUTF8Buf(buffer,count,tmpBuf^);
        utfWritten := tfsStream.Write(tmpBuf^,numBytes);
        if utfWritten <> numBytes then begin
          Result := 0; // to keep Delphi from complaining
          // To find out how much data was actually written (in term of Unicode
          // characters) we have to decode written data back to Unicode. Ouch.
          GetMem(uniBuf,count); // decoded data cannot use more space than original Unicode data
          try
            Result := UTF8BufToWideCharBuf(tmpBuf^,Result,uniBuf^,leftUTF8);
          finally FreeMem(uniBuf); end;
        end
        else // everything was written
          Result := count;
      finally FreeBuffer(tmpBuf); end;
    end
    else
      Result := tfsStream.Write(buffer,count);
  end
  else begin
    if Odd(count) then
      raise EGpTextFileStream.CreateFmtHelp(sCannotConvertOddNumberOfBytes,[FileName,count],hcTFCannotConvertOdd)
    else begin
      numChar := count div SizeOf(WideChar);
      tmpBuf := AllocBuffer(numChar);
      try
        numChar := WideCharToMultiByte(tfsCodePage,
          WC_COMPOSITECHECK or WC_DISCARDNS or WC_SEPCHARS or WC_DEFAULTCHAR,
          @buffer, numChar, tmpBuf, numChar, nil, nil);
        Win32Check(numChar <> 0,'Write');
        Result := tfsStream.Write(tmpBuf^,numChar) * SizeOf(WideChar);
      finally FreeBuffer(tmpBuf); end;
    end;
  end;
end; { TGpTextFileStream.Write }

{:Writes string to stream and terminates it with line delimiter (as set in
  constructor). If file is 8-bit, data is converted from Unicode according to
  code page specified in constructor.
  @param   ln String to be written.
  @returns True if string was written successfully.
  @raises  EGpTextFileStream if conversion from 8-bit to Unicode failes.
  @raises  EGpHugeFile on Windows errors.
}
function TGpTextFileStream.Writeln(const ln: WideString): boolean;
var
  ch: AnsiChar;
  wc: WideChar;
begin
  if ln <> '' then begin
    if not WriteString(ln) then begin
      Result := false;
      Exit;
    end;
  end;
  if IsUnicode then begin
    if cfUse2028 in tfsCreateFlags then begin
      wc := WideChar($2028);
      Result := (Write(wc,SizeOf(WideChar)) = SizeOf(WideChar));
    end
    else begin
      wc := WideChar($000D);
      Result := (Write(wc,SizeOf(WideChar)) = SizeOf(WideChar));
      if Result then begin
        wc := WideChar($000A);
        Result := (Write(wc,SizeOf(WideChar)) = SizeOf(WideChar));
      end;
    end;
  end
  else begin
    if cfUseLF in tfsCreateFlags then begin
      ch := Char($0D);
      Result := (tfsStream.Write(ch,SizeOf(Char)) = SizeOf(Char));
    end
    else begin
      ch := Char($0D);
      Result := (tfsStream.Write(ch,SizeOf(Char)) = SizeOf(Char));
      if Result then begin
        ch := Char($0A);
        Result := (tfsStream.Write(ch,SizeOf(Char)) = SizeOf(Char));
      end;
    end;
  end;
end; { TGpTextFileStream.Writeln }

{:Writes string to stream. If file is 8-bit, data is converted from Unicode
  according to code page specified in constructor.
  @param   ws String to be written.
  @returns True if string was written successfully.
  @raises  EGpTextFileStream if conversion from 8-bit to Unicode failes.
  @raises  EGpHugeFile on Windows errors.
}
function TGpTextFileStream.WriteString(const ws: WideString): boolean;
begin
  Result := (Write(ws[1],Length(ws)*SizeOf(WideChar)) = Length(ws)*SizeOf(WideChar));
end; { TGpTextFileStream.WriteString }

end.
